@@ -20,6 +20,7 @@
 #include <locale.h>
 #include <sys/types.h>
 #include <utime.h>
+#include <stdint.h>
 
 #if defined(SunOS)
 #include <limits.h>
@@ -72,15 +73,14 @@ extern int asprintf(char **str, const char *fmt, ...);
 #define DOEXIT exit
 #endif
 
-#define ERROR_STOP_ROTATION 1
-#define ERROR_CONTINUE_ROTATION 2
 
 struct logState {
-    char *fn;
-    struct tm lastRotated;	/* only tm_hour, tm_mday, tm_mon, tm_year are good! */
-    struct stat sb;
-    int doRotate;
-    LIST_ENTRY(logState) list;
+	char *fn;
+	struct tm lastRotated;	/* only tm_hour, tm_mday, tm_mon, tm_year are good! */
+	struct stat sb;
+	int doRotate;
+	int isUsed;	/* True if there is real log file in system for this state. */
+	LIST_ENTRY(logState) list;
 };
 
 struct logNames {
@@ -239,23 +239,17 @@ static void unescape(char *arg)
 }
 
 #define HASH_SIZE_MIN 64
-static int allocateHash(void)
+static int allocateHash(unsigned int hs)
 {
-	struct logInfo *log;
-	unsigned int hs;
 	int i;
 
-	hs = 0;
-
-	for (log = logs.tqh_first; log != NULL; log = log->list.tqe_next)
-		hs += log->numFiles;
-
-	hs *= 2;
-
 	/* Enforce some reasonable minimum hash size */
 	if (hs < HASH_SIZE_MIN)
 		hs = HASH_SIZE_MIN;
 
+	message(MESS_DEBUG, "Allocating hash table for state file, size %d entries\n",
+			hs);
+
 	states = calloc(hs, sizeof(struct logStates *));
 	if (states == NULL) {
 		message(MESS_ERROR, "could not allocate memory for "
@@ -306,6 +300,7 @@ static struct logState *newState(const char *fn)
 	}
 
 	new->doRotate = 0;
+	new->isUsed = 0;
 
 	memset(&new->lastRotated, 0, sizeof(new->lastRotated));
 	new->lastRotated.tm_hour = now.tm_hour;
@@ -525,6 +520,7 @@ static int removeLogFile(char *name, struct logInfo *log)
 static int compressLogFile(char *name, struct logInfo *log, struct stat *sb)
 {
     char *compressedName;
+    char *envInFilename;
     const char **fullCommand;
     struct utimbuf utim;
     int inFile;
@@ -599,6 +595,9 @@ static int compressLogFile(char *name, struct logInfo *log, struct stat *sb)
 		DOEXIT(1);
 	}
 
+	envInFilename = alloca(strlen("LOGROTATE_COMPRESSED_FILENAME=") + strlen(name) + 2);
+	sprintf(envInFilename, "LOGROTATE_COMPRESSED_FILENAME=%s", name);
+	putenv(envInFilename);
 	execvp(fullCommand[0], (void *) fullCommand);
 	DOEXIT(1);
     }
@@ -722,30 +721,149 @@ static int mailLog(struct logInfo *log, char *logFile, char *mailCommand,
 static int mailLogWrapper(char *mailFilename, char *mailCommand,
 			  int logNum, struct logInfo *log)
 {
-    /* if the log is compressed (and we're not mailing a
-     * file whose compression has been delayed), we need
-     * to uncompress it */
-    if ((log->flags & LOG_FLAG_COMPRESS) &&
-	!((log->flags & LOG_FLAG_DELAYCOMPRESS) &&
-	  (log->flags & LOG_FLAG_MAILFIRST))) {
-	if (mailLog(log, mailFilename, mailCommand,
-		    log->uncompress_prog, log->logAddress,
-		    log->files[logNum]))
-	    return 1;
-    } else {
-	if (mailLog(log, mailFilename, mailCommand, NULL,
-		    log->logAddress, mailFilename))
-	    return 1;
+	/* if the log is compressed (and we're not mailing a
+	* file whose compression has been delayed), we need
+	* to uncompress it */
+	if ((log->flags & LOG_FLAG_COMPRESS) && !(log->flags & LOG_FLAG_DELAYCOMPRESS)) {
+		if (mailLog(log, mailFilename, mailCommand,
+			log->uncompress_prog, log->logAddress,
+			(log->flags & LOG_FLAG_MAILFIRST) ? log->files[logNum] : mailFilename))
+			return 1;
+	} else {
+		if (mailLog(log, mailFilename, mailCommand, NULL,
+			log->logAddress, mailFilename))
+			return 1;
+	}
+	return 0;
+}
+
+/* Use a heuristic to determine whether stat buffer SB comes from a file
+   with sparse blocks.  If the file has fewer blocks than would normally
+   be needed for a file of its size, then at least one of the blocks in
+   the file is a hole.  In that case, return true.  */
+static int is_probably_sparse(struct stat const *sb)
+{
+#if defined(HAVE_STRUCT_STAT_ST_BLOCKS) && defined(HAVE_STRUCT_STAT_ST_BLKSIZE)
+	return (S_ISREG (sb->st_mode)
+          && sb->st_blocks < sb->st_size / sb->st_blksize);
+#else
+	return 0;
+#endif
+}
+
+#define MIN(a,b) ((a) < (b) ? (a) : (b))
+
+/* Return whether the buffer consists entirely of NULs.
+   Note the word after the buffer must be non NUL. */
+
+static inline int is_nul (void const *buf, size_t bufsize)
+{
+  char const *cbuf = buf;
+  char const *cp = buf;
+
+  /* Find the first nonzero *byte*, or the sentinel.  */
+  while (*cp++ == 0)
+    continue;
+
+  return cbuf + bufsize < cp;
+}
+
+static size_t full_write(int fd, const void *buf, size_t count)
+{
+  size_t total = 0;
+  const char *ptr = (const char *) buf;
+
+  while (count > 0)
+    {
+      size_t n_rw;
+	for (;;)
+	{
+		n_rw = write (fd, buf, count);
+		if (errno == EINTR)
+			continue;
+		else
+			break;
+	}
+	if (n_rw == (size_t) -1)
+		break;
+	if (n_rw == 0)
+		break;
+	total += n_rw;
+	ptr += n_rw;
+	count -= n_rw;
     }
-    return 0;
+
+  return total;
+}
+
+static int sparse_copy(int src_fd, int dest_fd, struct stat *sb,
+		       const char *saveLog, const char *currLog)
+{
+	int make_holes = is_probably_sparse(sb);
+	size_t max_n_read = SIZE_MAX;
+	int last_write_made_hole = 0;
+	off_t total_n_read = 0;
+	char buf[BUFSIZ + 1];
+
+	while (max_n_read) {
+		int make_hole = 0;
+
+		ssize_t n_read = read (src_fd, buf, MIN (max_n_read, BUFSIZ));
+		if (n_read < 0) {
+			if (errno == EINTR) {
+				continue;
+			}
+			message(MESS_ERROR, "error reading %s: %s\n",
+				currLog, strerror(errno));
+			return 0;
+		}
+
+		if (n_read == 0)
+			break;
+
+		max_n_read -= n_read;
+		total_n_read += n_read;
+
+		if (make_holes) {
+			/* Sentinel required by is_nul().  */
+			buf[n_read] = '\1';
+
+			if ((make_hole = is_nul(buf, n_read))) {
+				if (lseek (dest_fd, n_read, SEEK_CUR) < 0) {
+					message(MESS_ERROR, "error seeking %s: %s\n",
+						saveLog, strerror(errno));
+					return 0;
+				}
+			}
+		}
+
+		if (!make_hole) {
+			size_t n = n_read;
+			if (full_write (dest_fd, buf, n) != n) {
+				message(MESS_ERROR, "error writing to %s: %s\n",
+					saveLog, strerror(errno));
+				return 0;
+			}
+		}
+
+		last_write_made_hole = make_hole;
+	}
+
+	if (last_write_made_hole) {
+		if (ftruncate(dest_fd, total_n_read) < 0) {
+			message(MESS_ERROR, "error ftruncate %s: %s\n",
+			saveLog, strerror(errno));
+			return 0;
+		}
+	}
+
+	return 1;
 }
 
 static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
 			int flags)
 {
-    char buf[BUFSIZ];
     int fdcurr = -1, fdsave = -1;
-    ssize_t cnt;
 
     message(MESS_DEBUG, "copying %s to %s\n", currLog, saveLog);
 
@@ -753,7 +871,7 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
 	if ((fdcurr = open(currLog, ((flags & LOG_FLAG_COPY) ? O_RDONLY : O_RDWR) | O_NOFOLLOW)) < 0) {
 	    message(MESS_ERROR, "error opening %s: %s\n", currLog,
 		    strerror(errno));
-	    return ERROR_CONTINUE_ROTATION;
+	    return 1;
 	}
 #ifdef WITH_SELINUX
 	if (selinux_enabled) {
@@ -822,21 +940,10 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
 	    return 1;
 	}
 
-	while ((cnt = read(fdcurr, buf, sizeof(buf))) > 0) {
-	    if (write(fdsave, buf, cnt) != cnt) {
-		message(MESS_ERROR, "error writing to %s: %s\n",
-			saveLog, strerror(errno));
+	if (sparse_copy(fdcurr, fdsave, sb, saveLog, currLog) != 1) {
 		close(fdcurr);
 		close(fdsave);
 		return 1;
-	    }
-	}
-	if (cnt != 0) {
-	    message(MESS_ERROR, "error reading %s: %s\n",
-		    currLog, strerror(errno));
-	    close(fdcurr);
-	    close(fdsave);
-	    return 1;
 	}
     }
 
@@ -913,9 +1020,10 @@ int findNeedRotating(struct logInfo *log, int logNum, int force)
 	return 1;
     }
 
-    state = findState(log->files[logNum]);
-    state->doRotate = 0;
-    state->sb = sb;
+	state = findState(log->files[logNum]);
+	state->doRotate = 0;
+	state->sb = sb;
+	state->isUsed = 1;
 
 	if ((sb.st_mode & S_IFMT) == S_IFLNK) {
 	    message(MESS_DEBUG, "  log %s is symbolic link. Rotation of symbolic"
@@ -928,8 +1036,15 @@ int findNeedRotating(struct logInfo *log, int logNum, int force)
 	/* user forced rotation of logs from command line */
 	state->doRotate = 1;   
     }
+    else if (log->maxsize && sb.st_size > log->maxsize) {
+        state->doRotate = 1;
+    }
     else if (log->criterium == ROT_SIZE) {
 	state->doRotate = (sb.st_size >= log->threshhold);
+	if (!state->doRotate) {
+	message(MESS_DEBUG, "  log does not need rotating "
+		"(log size is below the 'size' threshold)\n");
+	}
     } else if (mktime(&state->lastRotated) - mktime(&now) > (25 * 3600)) {
         /* 25 hours allows for DST changes as well as geographical moves */
 	message(MESS_ERROR,
@@ -952,49 +1067,90 @@ int findNeedRotating(struct logInfo *log, int logNum, int force)
 			       ((mktime(&now) -
 				 mktime(&state->lastRotated)) >
 				(7 * 24 * 3600)));
+	    if (!state->doRotate) {
+	    message(MESS_DEBUG, "  log does not need rotating "
+		    "(log has been rotated at %d-%d-%d %d:%d, "
+		    "that is not week ago yet)\n", 1900 + state->lastRotated.tm_year,
+		    1 + state->lastRotated.tm_mon, state->lastRotated.tm_mday,
+		    state->lastRotated.tm_hour, state->lastRotated.tm_min);
+	    }
 	    break;
 	case ROT_HOURLY:
 	    state->doRotate = ((now.tm_hour != state->lastRotated.tm_hour) ||
 			    (now.tm_mday != state->lastRotated.tm_mday) ||
 			    (now.tm_mon != state->lastRotated.tm_mon) ||
 			    (now.tm_year != state->lastRotated.tm_year));
+	    if (!state->doRotate) {
+	    message(MESS_DEBUG, "  log does not need rotating "
+		    "(log has been rotated at %d-%d-%d %d:%d, "
+		    "that is not hour ago yet)\n", 1900 + state->lastRotated.tm_year,
+		    1 + state->lastRotated.tm_mon, state->lastRotated.tm_mday,
+		    state->lastRotated.tm_hour, state->lastRotated.tm_min);
+	    }
 	    break;
 	case ROT_DAYS:
 	    /* FIXME: only days=1 is implemented!! */
 	    state->doRotate = ((now.tm_mday != state->lastRotated.tm_mday) ||
 			    (now.tm_mon != state->lastRotated.tm_mon) ||
 			    (now.tm_year != state->lastRotated.tm_year));
+	    if (!state->doRotate) {
+	    message(MESS_DEBUG, "  log does not need rotating "
+		    "(log has been rotated at %d-%d-%d %d:%d, "
+		    "that is not day ago yet)\n", 1900 + state->lastRotated.tm_year,
+		    1 + state->lastRotated.tm_mon, state->lastRotated.tm_mday,
+		    state->lastRotated.tm_hour, state->lastRotated.tm_min);
+	    }
 	    break;
 	case ROT_MONTHLY:
 	    /* rotate if the logs haven't been rotated this month or
 	       this year */
 	    state->doRotate = ((now.tm_mon != state->lastRotated.tm_mon) ||
 			    (now.tm_year != state->lastRotated.tm_year));
+	    if (!state->doRotate) {
+	    message(MESS_DEBUG, "  log does not need rotating "
+		    "(log has been rotated at %d-%d-%d %d:%d, "
+		    "that is not month ago yet)\n", 1900 + state->lastRotated.tm_year,
+		    1 + state->lastRotated.tm_mon, state->lastRotated.tm_mday,
+		    state->lastRotated.tm_hour, state->lastRotated.tm_min);
+	    }
 	    break;
 	case ROT_YEARLY:
 	    /* rotate if the logs haven't been rotated this year */
 	    state->doRotate = (now.tm_year != state->lastRotated.tm_year);
+	    if (!state->doRotate) {
+	    message(MESS_DEBUG, "  log does not need rotating "
+		    "(log has been rotated at %d-%d-%d %d:%d, "
+		    "that is not year ago yet)\n", 1900 + state->lastRotated.tm_year,
+		    1 + state->lastRotated.tm_mon, state->lastRotated.tm_mday,
+		    state->lastRotated.tm_hour, state->lastRotated.tm_min);
+	    }
 	    break;
 	default:
 	    /* ack! */
 	    state->doRotate = 0;
 	    break;
 	}
-	if (log->minsize && sb.st_size < log->minsize)
+	if (log->minsize && sb.st_size < log->minsize) {
 	    state->doRotate = 0;
+	    message(MESS_DEBUG, "  log does not need rotating "
+		    "('misinze' directive is used and the log "
+		    "size is smaller than the minsize value");
+	}
+    }
+    else if (!state->doRotate) {
+	message(MESS_DEBUG, "  log does not need rotating "
+		"(log has been already rotated)");
     }
 
-    if (log->maxsize && sb.st_size > log->maxsize)
-        state->doRotate = 1;
-
     /* The notifempty flag overrides the normal criteria */
-    if (!(log->flags & LOG_FLAG_IFEMPTY) && !sb.st_size)
+    if (state->doRotate && !(log->flags & LOG_FLAG_IFEMPTY) && !sb.st_size) {
 	state->doRotate = 0;
+	message(MESS_DEBUG, "  log does not need rotating "
+		"(log is empty)");
+    }
 
     if (state->doRotate) {
 	message(MESS_DEBUG, "  log needs rotating\n");
-    } else {
-	message(MESS_DEBUG, "  log does not need rotating\n");
     }
 
     return 0;
@@ -1099,6 +1255,10 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 						j += 10; /* strlen("[0-9][0-9]") */
 					case 'm':
 					case 'd':
+					case 'H':
+					case 'M':
+					case 'S':
+					case 'V':
 						strncat(dext_pattern, "[0-9][0-9]",
 								sizeof(dext_pattern) - strlen(dext_pattern) - 1);
 						j += 10;
@@ -1250,9 +1410,10 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 	}
     }
 
+    /* adding 2 due to / and \0 being added by snprintf */
     rotNames->firstRotated =
 	malloc(strlen(rotNames->dirName) + strlen(rotNames->baseName) +
-	       strlen(fileext) + strlen(compext) + 30);
+	       strlen(fileext) + strlen(compext) + DATEEXT_LEN + 2 );
 
     if (log->flags & LOG_FLAG_DATEEXT) {
 	/* glob for compressed files with our pattern
@@ -1411,10 +1572,9 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 	}
 
     /* if the last rotation doesn't exist, that's okay */
-    if (!debug && rotNames->disposeName
-	&& access(rotNames->disposeName, F_OK)) {
+    if (rotNames->disposeName && access(rotNames->disposeName, F_OK)) {
 	message(MESS_DEBUG,
-		"log %s doesn't exist -- won't try to " "dispose of it\n",
+		"log %s doesn't exist -- won't try to dispose of it\n",
 		rotNames->disposeName);
 	free(rotNames->disposeName);
 	rotNames->disposeName = NULL;
@@ -1501,7 +1661,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 			if (!ACL_NOT_WELL_SUPPORTED(errno)) {
 				message(MESS_ERROR, "getting file ACL %s: %s\n",
 					log->files[logNum], strerror(errno));
-				hasErrors |= 1;
+				hasErrors = 1;
 			}
 		}
 #endif /* WITH_ACL */
@@ -1517,12 +1677,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 			message(MESS_ERROR, "failed to rename %s to %s: %s\n",
 				log->files[logNum], tmpFilename,
 				strerror(errno));
-				if (errno == ENOENT) {
-					hasErrors |= ERROR_CONTINUE_ROTATION;
-				}
-				else {
-					hasErrors |= 1;
-				}
+				hasErrors = 1;
 			}
 		}
 		else {
@@ -1533,7 +1688,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 				message(MESS_ERROR, "failed to rename %s to %s: %s\n",
 					log->files[logNum], tmpFilename,
 					strerror(errno));
-					hasErrors |= 1;
+					hasErrors = 1;
 			}
 	    }
 
@@ -1587,7 +1742,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 			}
 #endif
 			if (fd < 0)
-				hasErrors |= 1;
+				hasErrors = 1;
 			else {
 				close(fd);
 			}
@@ -1605,7 +1760,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 	if (!hasErrors
 	    && log->flags & (LOG_FLAG_COPYTRUNCATE | LOG_FLAG_COPY)
 		&& !(log->flags & LOG_FLAG_TMPFILENAME)) {
-	    hasErrors |=
+	    hasErrors =
 		copyTruncate(log->files[logNum], rotNames->finalName,
 			     &state->sb, log->flags);
 	}
@@ -1680,7 +1835,6 @@ int rotateLogSet(struct logInfo *log, int force)
     int i, j;
     int hasErrors = 0;
     int logHasErrors[log->numFiles];
-	int logHasScriptErrors[log->numFiles];
     int numRotated = 0;
     struct logState **state;
     struct logNames **rotNames;
@@ -1752,7 +1906,6 @@ int rotateLogSet(struct logInfo *log, int force)
     for (i = 0; i < log->numFiles; i++) {
 	logHasErrors[i] = findNeedRotating(log, i, force);
 	hasErrors |= logHasErrors[i];
-	logHasScriptErrors[i] = 0;
 
 	/* sure is a lot of findStating going on .. */
 	if ((findState(log->files[i]))->doRotate)
@@ -1821,8 +1974,7 @@ int rotateLogSet(struct logInfo *log, int force)
 				"error running non-shared prerotate script "
 				"for %s of '%s'\n", log->files[j], log->pattern);
 		    }
-		    logHasScriptErrors[j] = 1;
-			logHasErrors[j] = 1;
+		    logHasErrors[j] = 1;
 		    hasErrors = 1;
 		}
 	    }
@@ -1831,7 +1983,8 @@ int rotateLogSet(struct logInfo *log, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-	    if (!logHasScriptErrors[i] && !(logHasErrors[i] & ERROR_STOP_ROTATION)) {
+	    if (! ( (logHasErrors[i] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) ) {
 		logHasErrors[i] |=
 		    rotateSingleLog(log, i, state[i], rotNames[i]);
 		hasErrors |= logHasErrors[i];
@@ -1840,7 +1993,7 @@ int rotateLogSet(struct logInfo *log, int force)
 
 	if (log->post
 		&& (!(
-			(((logHasErrors[j] & ERROR_STOP_ROTATION) || !state[j]->doRotate) && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+			((logHasErrors[j] || !state[j]->doRotate) && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
 			|| (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS))
 		))
 	) {
@@ -1859,8 +2012,7 @@ int rotateLogSet(struct logInfo *log, int force)
 				"error running non-shared postrotate script "
 				"for %s of '%s'\n", log->files[j], log->pattern);
 		    }
-		    logHasScriptErrors[j] = 1;
-			logHasErrors[j] = 1;
+		    logHasErrors[j] = 1;
 		    hasErrors = 1;
 		}
 	    }
@@ -1869,7 +2021,8 @@ int rotateLogSet(struct logInfo *log, int force)
 	for (i = j;
 	     ((log->flags & LOG_FLAG_SHAREDSCRIPTS) && i < log->numFiles)
 	     || (!(log->flags & LOG_FLAG_SHAREDSCRIPTS) && i == j); i++) {
-		if (!logHasScriptErrors[i] && !(logHasErrors[i] & ERROR_STOP_ROTATION)) {
+	    if (! ( (logHasErrors[i] && !(log->flags & LOG_FLAG_SHAREDSCRIPTS))
+		   || (hasErrors && (log->flags & LOG_FLAG_SHAREDSCRIPTS)) ) ) {
 		logHasErrors[i] |=
 		    postrotateSingleLog(log, i, state[i], rotNames[i]);
 		hasErrors |= logHasErrors[i];
@@ -1924,6 +2077,8 @@ static int writeState(char *stateFilename)
 	int fdsave;
 	struct stat sb;
 	char *tmpFilename = NULL;
+	struct tm now = *localtime(&nowSecs);
+	time_t now_time, last_time;
 
 	tmpFilename = malloc(strlen(stateFilename) + 5 );
 	if (tmpFilename == NULL) {
@@ -2028,9 +2183,22 @@ static int writeState(char *stateFilename)
 	if (bytes < 0)
 		error = bytes;
 
+#define SECONDS_IN_YEAR 31556926
+
 	for (i = 0; i < hashSize && error == 0; i++) {
 		for (p = states[i]->head.lh_first; p != NULL && error == 0;
 				p = p->list.le_next) {
+
+			/* Skip states which are not used for more than a year. */
+			now_time = mktime(&now);
+			last_time = mktime(&p->lastRotated);
+			if (!p->isUsed && difftime(now_time, last_time) > SECONDS_IN_YEAR) {
+				message(MESS_DEBUG, "Removing %s from state file, "
+					"because it does not exist and has not been rotated for one year\n",
+					p->fn);
+				continue;
+			}
+
 			error = fputc('"', f) == EOF;
 			for (chptr = p->fn; *chptr && error == 0; chptr++) {
 				switch (*chptr) {
@@ -2114,23 +2282,27 @@ static int readState(char *stateFilename)
 
     error = stat(stateFilename, &f_stat);
 
-    if ((error && errno == ENOENT) || (!error && f_stat.st_size == 0)) {
-	/* create the file before continuing to ensure we have write
-	   access to the file */
-	f = fopen(stateFilename, "w");
-	if (!f) {
-	    message(MESS_ERROR, "error creating state file %s: %s\n",
-		    stateFilename, strerror(errno));
-	    return 1;
+	if ((error && errno == ENOENT) || (!error && f_stat.st_size == 0)) {
+		/* create the file before continuing to ensure we have write
+		access to the file */
+		f = fopen(stateFilename, "w");
+		if (!f) {
+			message(MESS_ERROR, "error creating state file %s: %s\n",
+				stateFilename, strerror(errno));
+			return 1;
+		}
+		fprintf(f, "logrotate state -- version 2\n");
+		fclose(f);
+
+		if (allocateHash(64) != 0)
+			return 1;
+
+		return 0;
+	} else if (error) {
+		message(MESS_ERROR, "error stat()ing state file %s: %s\n",
+			stateFilename, strerror(errno));
+		return 1;
 	}
-	fprintf(f, "logrotate state -- version 2\n");
-	fclose(f);
-	return 0;
-    } else if (error) {
-	message(MESS_ERROR, "error stat()ing state file %s: %s\n",
-		stateFilename, strerror(errno));
-	return 1;
-    }
 
     f = fopen(stateFilename, "r");
     if (!f) {
@@ -2154,6 +2326,13 @@ static int readState(char *stateFilename)
 	return 1;
     }
 
+	/* Try to estimate how many state entries we have in the state file.
+	 * We expect single entry to have around 80 characters (Of course this is
+	 * just an estimation). During the testing I've found out that 200 entries
+	 * per single hash entry gives good mem/performance ratio. */
+	if (allocateHash(f_stat.st_size / 80 / 200) != 0)
+		return 1;
+
     line++;
 
     while (fgets(buf, sizeof(buf) - 1, f)) {
@@ -2183,7 +2362,7 @@ static int readState(char *stateFilename)
 	}
 
 	/* Hack to hide earlier bug */
-	if ((year != 1900) && (year < 1996 || year > 2100)) {
+	if ((year != 1900) && (year < 1970 || year > 2100)) {
 	    message(MESS_ERROR,
 		    "bad year %d for file %s in state file %s\n", year,
 		    argv[0], stateFilename);
@@ -2273,6 +2452,8 @@ int main(int argc, const char **argv)
 {
     int force = 0;
     char *stateFile = STATEFILE;
+    char *logFile = NULL;
+    FILE *logFd = 0;
     int rc = 0;
     int arg;
     const char **files;
@@ -2290,6 +2471,7 @@ int main(int argc, const char **argv)
 	 "Path of state file",
 	 "statefile"},
 	{"verbose", 'v', 0, 0, 'v', "Display messages during rotation"},
+	{"log", 'l', POPT_ARG_STRING, &logFile, 'l', "Log file or 'syslog' to log to syslog"},
 	{"version", '\0', POPT_ARG_NONE, NULL, 'V', "Display version information"},
 	POPT_AUTOHELP {0, 0, 0, 0, 0}
     };
@@ -2309,6 +2491,20 @@ int main(int argc, const char **argv)
 	case 'v':
 	    logSetLevel(MESS_DEBUG);
 	    break;
+	case 'l':
+		if (strcmp(logFile, "syslog") == 0) {
+			logToSyslog(1);
+		}
+		else {
+			logFd = fopen(logFile, "w");
+			if (!logFd) {
+				message(MESS_ERROR, "error opening log file %s: %s\n",
+					logFile, strerror(errno));
+				break;
+			}
+			logSetMessageFile(logFd);
+		}
+		break;
 	case 'V':
 	    fprintf(stderr, "logrotate %s\n", VERSION);
 	    poptFreeContext(optCon);
@@ -2350,9 +2546,6 @@ int main(int argc, const char **argv)
     poptFreeContext(optCon);
     nowSecs = time(NULL);
 
-	if (allocateHash() != 0)
-		return 1;
-
 	if (readState(stateFile))
 		exit(1);
 
